@@ -21,7 +21,6 @@
 #ifndef __PINENTRY_QT_ACCESSIBILITY_H__
 #define __PINENTRY_QT_ACCESSIBILITY_H__
 
-class QLabel;
 class QString;
 class QWidget;
 
@@ -36,10 +35,6 @@ void setDescription(QWidget *w, const QString &text);
    QT_NO_ACCESSIBILITY is defined. */
 void setName(QWidget *w, const QString &text);
 
-/* Selects the text displayed by the label. Only QLabel with text format
-   Qt::PlainText or Qt::RichText are supported. */
-void selectLabelText(QLabel *label);
-
 } // namespace Accessibility
 
 #endif // __PINENTRY_QT_ACCESSIBILITY_H__
